package org.artifactory.projects.interceptor;

import org.artifactory.config.ConfigurationChangesInterceptor;
import org.artifactory.config.InternalCentralConfigService;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.repo.RepoDescriptor;
import org.artifactory.projects.ProjectsService;
import org.artifactory.projects.exception.AccessClientStateException;
import org.artifactory.rest.exception.BadRequestException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.Map;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Handles repository removal from Projects (managed in Access) for local, remote and virtual repositories
 *
 * @author andreik
 */
@Component
public class ProjectRepoConfigurationInterceptor implements ConfigurationChangesInterceptor {

    private static final Logger log = LoggerFactory.getLogger(ProjectRepoConfigurationInterceptor.class);

    private final InternalCentralConfigService internalCentralConfigService;
    private final ProjectsService projectsService;

    @Autowired
    public ProjectRepoConfigurationInterceptor(InternalCentralConfigService internalCentralConfigService,
            ProjectsService projectsService) {
        this.internalCentralConfigService = internalCentralConfigService;
        this.projectsService = projectsService;
    }

    @Override
    public void onBeforeSave(CentralConfigDescriptor newDescriptor) {
        CentralConfigDescriptor oldDescriptor = internalCentralConfigService.getDescriptor();
        if (oldDescriptor != null && newDescriptor != null) {
            // Cannot be duplicates of repositories in descriptor.
            Map<String, ? extends RepoDescriptor> newRepos = createConcatedRepositoriesMap(newDescriptor);
            Map<String, ? extends RepoDescriptor> oldRepos = createConcatedRepositoriesMap(oldDescriptor);
            removeDeletedRepositoriesFromProjects(newRepos, oldRepos);
        }
    }

    private Map<String, ? extends RepoDescriptor> createConcatedRepositoriesMap(CentralConfigDescriptor descriptor) {
        try {
            return Stream.of(descriptor.getLocalAndFederatedRepositoriesMap(),
                            descriptor.getRemoteRepositoriesMap(),
                            descriptor.getVirtualRepositoriesMap())
                    .flatMap(map -> map.entrySet().stream())
                    .collect(Collectors.toMap(
                            Map.Entry::getKey,
                            Map.Entry::getValue));
        } catch (IllegalStateException e) {
            log.debug("Duplicated Repo in descriptor: {}", descriptor);
            log.error("Failed mapping all repos in descriptor ", e);
            throw new BadRequestException("Duplicated repo detected in the descriptor check the logs.");
        }
    }

    private void removeDeletedRepositoriesFromProjects(Map<String, ? extends RepoDescriptor> newRepos,
            Map<String, ? extends RepoDescriptor> oldRepos) {
        if (newRepos != null && oldRepos != null) {
            for (Map.Entry<String, ? extends RepoDescriptor> repoKeyEntry : oldRepos.entrySet()) {
                String repoKey = repoKeyEntry.getKey();
                if (!newRepos.containsKey(repoKey)) {
                    try {
                        // for every project remove the repo
                        log.debug("Repo '{}' from type '{}' was deleted, removing from Access", repoKey,
                                repoKeyEntry.getValue().getType());
                        projectsService.removeRepositoryFromCurrentProject(repoKey);
                    } catch (AccessClientStateException e) {
                        log.warn("Failed to remove repo '{}' from a project association: {}. Will be handled by " +
                                "SyncResourcesWithAccessJob", repoKey, e.getMessage());
                    } catch (Exception e) {
                        log.error("Failed to remove repo '{}' from a project association: {}", repoKey, e.getMessage());
                        log.debug("Failed to remove repo '{}' from a project association: {}", repoKey, e.getMessage(),
                                e);
                    }
                }
            }
        }
    }
}
