package org.artifactory.util;

import lombok.extern.slf4j.Slf4j;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.security.SecurityService;
import org.artifactory.factory.InfoFactoryHolder;
import org.artifactory.security.SecurityInfo;
import org.artifactory.security.exceptions.LoginDisabledException;
import org.springframework.security.authentication.LockedException;
import org.springframework.security.web.authentication.www.BasicAuthenticationEntryPoint;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;

/**
 * @author Mor Merhav
 */
@Slf4j
public final class SecurityUtils {
    private static SecurityService securityService;

    private SecurityUtils() {

    }

    public static boolean ensureLoginAttemptAllowed(HttpServletRequest request, HttpServletResponse response,
            String loginIdentifier, BasicAuthenticationEntryPoint authenticationEntryPoint) throws IOException {
        try {
            long accessTime = HttpUtils.getSessionAccessTime(request);
            String remoteClientAddress = HttpUtils.getRemoteClientAddress(request);

            if (loginIdentifier == null) {
                return true;
            }
            if (loginIdentifier.isEmpty()) {
                getSecurityService().ensureSessionIsNotLocked(loginIdentifier);
                getSecurityService().ensureSessionShouldNotBeDelayed(loginIdentifier);
            } else {
                getSecurityService().ensureUserIsNotLocked(loginIdentifier);
                getSecurityService().ensureLoginShouldNotBeDelayed(loginIdentifier);
            }
            getSecurityService().updateUserLastAccess(loginIdentifier, remoteClientAddress, accessTime);
        } catch (LockedException | LoginDisabledException e) {
            log.debug("{}, cause: {}", e.getMessage(), e);
            authenticationEntryPoint.commence(request, response, e);
            return false;
        }
        return true;
    }

    /**
     * This function is used for encrypting sensitive data that should not be displayed as part of the security
     * descriptor (that should be removed in the future)
     *
     * @return SecurityInfo with encrypted values of sensitive data
     */
    public static SecurityInfo getEncryptSecurityInfo() {
        SecurityInfo securityInfo = getSecurityService().getSecurityData(false);
        SecurityInfo encryptSecurityInfo = InfoFactoryHolder.get().createSecurityInfo(
                securityInfo.getUsers(),
                securityInfo.getGroups(),
                securityInfo.getRepoAcls(),
                securityInfo.getBuildAcls(),
                securityInfo.getReleaseBundleAcls()
        );
        encryptSecurityInfo.setVersion(securityInfo.getVersion());
        return encryptSecurityInfo;
    }

    private static SecurityService getSecurityService() {
        if (securityService == null) {
            securityService = ContextHelper.get().beanForType(SecurityService.class);
        }
        return securityService;
    }
}
