/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.util;

import com.thoughtworks.xstream.XStream;
import com.thoughtworks.xstream.security.ArrayTypePermission;
import com.thoughtworks.xstream.security.NoTypePermission;
import com.thoughtworks.xstream.security.NullPermission;
import com.thoughtworks.xstream.security.PrimitiveTypePermission;
import lombok.experimental.UtilityClass;
import org.artifactory.exception.DocumentBuildException;
import org.jdom2.Document;
import org.jdom2.JDOMException;
import org.jdom2.input.SAXBuilder;
import org.jdom2.input.sax.XMLReaders;
import org.jdom2.output.Format;
import org.jdom2.output.XMLOutputter;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.StringReader;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * @author yoavl
 */
@UtilityClass
public final class XmlUtils {
    private static final String DOM_ERROR_MESSAGE = "Failed to build dom document";
    private static final String VALIDATION_FEATURE = "http://xml.org/sax/features/validation";
    private static final String DISALLOW_DOCTYPE_DECL_FEATURE = "http://apache.org/xml/features/disallow-doctype-decl";
    private static final String LOAD_DTD_GRAMMAR_FEATURE =
            "http://apache.org/xml/features/nonvalidating/load-dtd-grammar";
    private static final String LOAD_EXTERNAL_DTD_FEATURE =
            "http://apache.org/xml/features/nonvalidating/load-external-dtd";
    private static final String EXTERNAL_GENERAL_ENTITIES_FEATURE =
            "http://xml.org/sax/features/external-general-entities";
    private static final String EXTERNAL_PARAMETER_ENTITIES_FEATURE =
            "http://xml.org/sax/features/external-parameter-entities";
    private static final List<String> featuresToDisable = List.of(
            VALIDATION_FEATURE,
            LOAD_DTD_GRAMMAR_FEATURE,
            LOAD_EXTERNAL_DTD_FEATURE,
            EXTERNAL_GENERAL_ENTITIES_FEATURE,
            EXTERNAL_PARAMETER_ENTITIES_FEATURE);

    public static String outputString(Document doc) {
        return new XMLOutputter().outputString(doc);
    }

    public static String outputString(Format format, Document doc) {
        return new XMLOutputter(format).outputString(doc);
    }

    public static Document parse(String xmlContent) {
        SAXBuilder sb = createSaxBuilder();
        try {
            return sb.build(new StringReader(xmlContent));
        } catch (Exception e) {
            throw new DocumentBuildException(e.getMessage(), e);
        }
    }

    public static Document parseStrict(String xmlContent) {
        SAXBuilder sb = createStrictSaxBuilder();
        try {
            return sb.build(new StringReader(xmlContent));
        } catch (Exception e) {
            throw new DocumentBuildException(e.getMessage(), e);
        }
    }

    /**
     * Do not touch this method before you read this article :http://x-stream.github.io/security.html#example,
     * Merge your changes after a code review and a consultation with the security team.
     * This method stiffen the allowed types regarding the loading from xml by xstream.
     */
    public static void hardenXStream(XStream xstream) {
        // Clear out existing permissions and start a whitelist
        xstream.addPermission(NoTypePermission.NONE);

        // Allow null as type.
        xstream.addPermission(NullPermission.NULL);

        // Allow any primitive type and its boxed counterpart (excluding void).
        xstream.addPermission(PrimitiveTypePermission.PRIMITIVES);

        // Allow any array type, you can create an array of forbidden type
        // You cannot store objects inside the array.
        xstream.addPermission(ArrayTypePermission.ARRAYS);

        // Allow the types that were mentioned in the arrays.
        // When we set Map.class for example, we allowed only the Map objects only and not the successors Map.
        xstream.allowTypes(new Class[] {List.class, LinkedList.class, ArrayList.class});
        xstream.allowTypes(new Class[] {Set.class, HashSet.class, LinkedHashSet.class});
        xstream.allowTypes(new Class[] {Map.class, HashMap.class});
        xstream.allowTypes(new Class[] {String.class});

        // Allow the types under org.artifactory and org.jfrog
        xstream.allowTypesByWildcard(new String[] {"org.artifactory.**", "org.jfrog.**"});
    }

    public static Document parse(File input) {
        SAXBuilder sb = createSaxBuilder();
        try {
            return sb.build(new BufferedReader(
                    new InputStreamReader(new FileInputStream(input), StandardCharsets.UTF_8)));
        } catch (Exception e) {
            throw new DocumentBuildException(DOM_ERROR_MESSAGE, e);
        }
    }

    public static Document parse(InputStream in) {
        SAXBuilder sb = createSaxBuilder();
        try {
            return sb.build(in);
        } catch (Exception e) {
            throw new DocumentBuildException(DOM_ERROR_MESSAGE, e);
        }
    }

    public static SAXBuilder createSaxBuilder() {
        return createSaxBuilder(false);
    }

    public static SAXBuilder createStrictSaxBuilder() {
        return createSaxBuilder(true);
    }

    public static XMLReader createStrictXmlReader() throws JDOMException, SAXException {
        return createXmlReader(true);
    }

    public static XMLReader createXmlReader() throws JDOMException, SAXException {
        return createXmlReader(false);
    }

    public static XMLReader createXmlReader(boolean strict)
            throws JDOMException, SAXException {
        SAXBuilder sb = createSaxBuilder(false);
        XMLReader reader = sb.getXMLReaderFactory().createXMLReader();
        for (String feature : featuresToDisable) {
            reader.setFeature(feature, false);
        }
        if (strict) {
            reader.setFeature(DISALLOW_DOCTYPE_DECL_FEATURE, true);
        }
        return reader;
    }

    private static SAXBuilder createSaxBuilder(boolean strict) {
        SAXBuilder sb = new SAXBuilder();
        //don't validate and don't load dtd
        sb.setXMLReaderFactory(XMLReaders.NONVALIDATING);
        for (String feature : featuresToDisable) {
            sb.setFeature(feature, false);
        }
        if (strict) {
            sb.setFeature(DISALLOW_DOCTYPE_DECL_FEATURE, true);
        }
        return sb;
    }
}