package org.artifactory.core.api.ph.virtual;

import com.google.common.collect.Maps;
import org.artifactory.common.repo.PackageType;
import org.artifactory.common.resource.BaseResource;
import org.artifactory.confprop.ConfProps;
import org.artifactory.core.api.ph.exception.PackageException;
import org.artifactory.core.api.ph.exception.PackageVirtualTraversalException;
import org.artifactory.core.api.ph.resource.PackageUnfoundedResource;
import org.artifactory.core.api.ph.service.PackageArtifactService;
import org.artifactory.core.api.ph.service.PackageRepoService;
import org.artifactory.core.api.ph.service.PackageUploadService;
import org.artifactory.core.api.ph.virtual.resolution.PackageMetadataRow;
import org.artifactory.core.api.ph.virtual.resolution.PackageMetadataVersionsSet;
import org.artifactory.core.api.ph.virtual.resolution.PackageRepo;
import org.artifactory.core.api.ph.virtual.resolution.PackageRequestContext;
import org.artifactory.core.api.ph.virtual.resolution.PackageVirtualController;
import org.artifactory.core.api.ph.virtual.resolution.PackageVirtualResolver;
import org.artifactory.core.api.ph.virtual.resolution.VirtualResolutionStrategyType;
import org.artifactory.core.api.ph.virtual.resolution.VirtualResolverState;
import org.jetbrains.annotations.NotNull;
import org.jfrog.common.concurrency.ConflictsGuard;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nullable;
import java.io.InputStream;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.TimeUnit;
import java.util.function.BiFunction;

import static org.apache.http.HttpStatus.SC_FORBIDDEN;
import static org.apache.http.HttpStatus.SC_NOT_FOUND;

/**
 * Virtual repository, it is the engine/heart of the virtual repository resolution,
 * it provides the following characteristics
 * 1. Artifact resolution
 * 2. metadata merging
 * 3. virtual caching
 * 4. Priority resolution for both binary and metadata
 */
public class VirtualRepoTraversal {
    private static final Logger log = LoggerFactory.getLogger(VirtualRepoTraversal.class);
    protected static final String MERGE_LOCK_TIMEOUT_PROP_GLOBAL_KEY = "virtual.repo.metadata.merge.lock.timeout.sec";
    protected static final String MERGE_LOCK_TIMEOUT_PROP_PACKAGE_KEY =
            "virtual.repo.metadata.merge.lock.timeout.sec.%1$s";
    protected static final String MERGE_LOCK_TIMEOUT_PROP_REPO_KEY =
            "virtual.repo.metadata.merge.lock.timeout.sec.%1$s.%2$s";

    protected static final int DEFAULT_MERGE_LOCK_TIMEOUT_SEC = 60;

    private final PackageRepoService repoService;
    private final PackageArtifactService downloadService;
    private final PackageUploadService uploadService;
    private final ConflictsGuard<String> guard;
    private final ConfProps confProps;

    public VirtualRepoTraversal(PackageRepoService repoService, PackageArtifactService downloadService,
            PackageUploadService uploadService, ConflictsGuard<String> guard,
            ConfProps confProps) {
        this.repoService = repoService;
        this.downloadService = downloadService;
        this.uploadService = uploadService;
        this.guard = guard;
        this.confProps = confProps;
    }

    /**
     * VirtualVisitor one and only entry point.
     *
     * @param context             Request context that contains all the information required for the resolution/merging
     * @param controller          controller that defined the resolving/merging behaviour
     * @param getResourceFromRepo BIFunction that get requestContext and real repo as parameters, and returning a
     *                            resource.
     * @return resource that contains the resolution result which is item info
     */
    public BaseResource resolve(PackageRequestContext context, PackageVirtualController<?> controller,
            BiFunction<PackageRequestContext, PackageRepo, BaseResource> getResourceFromRepo) throws
            PackageVirtualTraversalException {
        log.debug("Consulting the virtual repo download strategy");
        PackageRequestContext cachePath = controller.getCachePath(context);
        BaseResource cachedResource = getInfoFromCache(cachePath, controller);
        if (isAvailable(cachedResource)) {
            return cachedResource;
        } else {
            BaseResource searchableResource =
                    getInfoFromSearchableRepositories(context, controller, getResourceFromRepo);
            cleanCacheIfNecessary(cachedResource, searchableResource);
            log.debug("Returning resource as found in the aggregated repositories");
            return searchableResource;
        }
    }

    private void cleanCacheIfNecessary(BaseResource cachedResource, BaseResource searchableResource) {
        // If cache is found but not available (expired/blocked) and searchableResource not exist clean the cache
        if (searchableResource != null && !searchableResource.isFound() && isFoundButNotAvailable(cachedResource)) {
            log.debug("Resource was not found but is still cached - removing from the cache");
            repoService.undeployArtifact(cachedResource.getRepoKey(), cachedResource.getPath());
        }
    }

    private boolean isFoundButNotAvailable(BaseResource cachedResource) {
        return cachedResource != null && (cachedResource.isExpired() || cachedResource.isBlocked());
    }

    private boolean isAvailable(BaseResource cachedResource) {
        // 1: Found resource
        // 2: Not expired
        // 3: Not blocked
        return cachedResource != null && cachedResource.isFound() && !cachedResource.isExpired() &&
                !cachedResource.isBlocked();
    }

    @Nullable
    private BaseResource getInfoFromCache(PackageRequestContext context, PackageVirtualController<?> controller)
            throws PackageVirtualTraversalException {
        VirtualResolutionStrategyType virtualResolutionStrategy =
                controller.getVirtualResolutionStrategy(context);
        if (VirtualResolutionStrategyType.MERGER == virtualResolutionStrategy) {
            try {
                log.debug("Trying to retrieve resource info from the local storage: {}/{}",
                        context.getRepoKey(), context.getPath());
                return downloadService.getResourceFromVirtualCache(context.getRepoKey(), context.getPath());
            } catch (PackageException e) {
                throw new PackageVirtualTraversalException(e);
            }
        }
        return null;
    }

    /**
     * Identify the resolution type:
     * 1. simple
     * 2. generated
     *
     * @param context    request context
     * @param controller package controller to control the merging or resolution process
     * @return Artifact resource
     */
    private BaseResource getInfoFromSearchableRepositories(PackageRequestContext context,
            PackageVirtualController<?> controller,
            BiFunction<PackageRequestContext, PackageRepo, BaseResource> getResourceFromRepo) {
        log.debug("Searching for info in aggregated repositories");
        try {
            List<? extends PackageRepo> repositories =
                    repoService.getSubRepositories(context.getRepoKey(), context.getPath(),
                            shouldFilterCacheRepos(controller, context));
            if (repositories == null || repositories.isEmpty()) {
                log.debug("Unable to find aggregated repositories to search within - returning un-found resource");
                return newUnfoundResource(context);
            }
            switch (controller.getVirtualResolutionStrategy(context)) {
                case RESOLVER: {
                    return searchInRepositories(context, controller, repositories, getResourceFromRepo);
                }
                case MERGER: {
                    return getMergedInfoFromSearchableRepositories(context, controller, repositories,
                            getResourceFromRepo);
                }
                default:
                    return newUnfoundResource(context);
            }
        } catch (PackageException e) {
            log.debug("Virtual repo: {} not exists ", context.getRepoKey(), e);
            return newUnfoundResource(context);
        }
    }

    private boolean shouldFilterCacheRepos(PackageVirtualController<?> controller, PackageRequestContext context) {
        return !controller.getVirtualResolutionStrategy(context).equals(VirtualResolutionStrategyType.RESOLVER) ||
                !controller.getResolverState().equals(VirtualResolverState.returnCurrentValue);
    }

    /**
     * Get the merged info from the cache or recalculate it.
     * using lock on the overriding repoKey/path
     *
     * @param context      the original repoKey/path
     * @param controller   the package controller that handle the merge
     * @param repositories the repositories contained in the virtual repository
     * @return the found resource or un-found resource
     */
    private BaseResource getMergedInfoFromSearchableRepositories(PackageRequestContext context,
            PackageVirtualController<?> controller, List<? extends PackageRepo> repositories,
            BiFunction<PackageRequestContext, PackageRepo, BaseResource> getResourceFromRepo) throws
            PackageVirtualTraversalException {
        PackageRequestContext cachePath = controller.getCachePath(context);
        try {
            String key = cachePath.getRepoKey() + "/" + cachePath.getPath();
            boolean locked = guard.tryToLock(key,
                    resolveMergeLockTimeoutSec(repoService.getPackageType(cachePath.getRepoKey()),
                            cachePath.getRepoKey()), TimeUnit.SECONDS);
            if (locked) {
                try {
                    BaseResource cachedResource = getInfoFromCache(cachePath, controller);
                    if (isAvailable(cachedResource)) {
                        return cachedResource;
                    } else {
                        return searchInRepositories(context, (PackageVirtualController<?>) controller, repositories,
                                cachePath, getResourceFromRepo);
                    }
                } finally {
                    guard.unlock(key);
                }
            } else {
                return getInfoFromCache(cachePath, controller);
            }
        } catch (InterruptedException e) {
            Thread.currentThread().interrupt();
            log.error("Failed to wait for merging virtual metadata, repoKey: {}, path: {}",
                    cachePath.getRepoKey(), cachePath.getPath(), e);
            BaseResource cachedResource = getInfoFromCache(cachePath, controller);
            if (isAvailable(cachedResource)) {
                return cachedResource;
            } else {
                return newUnfoundResource(context);
            }
        }
    }

    /**
     * Return the configurable merge lock timeout for the repo or the global default
     * {@code VirtualVisitor#mergeLockTimeoutSec}
     */
    private int resolveMergeLockTimeoutSec(PackageType packageType, String repoKey) {
        return confProps.getIntValue(String.format(MERGE_LOCK_TIMEOUT_PROP_REPO_KEY, packageType, repoKey),
                confProps.getIntValue(String.format(MERGE_LOCK_TIMEOUT_PROP_PACKAGE_KEY, packageType),
                        confProps.getIntValue(MERGE_LOCK_TIMEOUT_PROP_GLOBAL_KEY, DEFAULT_MERGE_LOCK_TIMEOUT_SEC)));
    }

    /**
     * Iterate over a list of repos, and return a merged resource to the given PackageVirtualController. The
     * order of searching is: local repos, and remote repos considering priority resolution.
     * The merged resource will be cached in the virtual cache.
     * Use this method in order to merge metadata files
     *
     * @param context             Original download request context
     * @param repositories        - sub repositories of a virtual repository - ordered by prioritised locals ->
     *                            prioritised remotes, locals -> remotes.
     * @param cachePath           path to store the merged resource under the virtual cache
     * @param getResourceFromRepo BIFunction that get requestContext and real repo as parameters, and returning a
     *                            resource.
     */
    private <T extends PackageMetadataRow> BaseResource searchInRepositories(PackageRequestContext context,
            PackageVirtualController<T> controller, List<? extends PackageRepo> repositories,
            PackageRequestContext cachePath,
            BiFunction<PackageRequestContext, PackageRepo, BaseResource> getResourceFromRepo) {
        PackageUnfoundedResource forbidden;
        Map<String, PackageMetadataVersionsSet<T>> rows = Maps.newHashMap();
        try {
            for (PackageRepo repo : repositories) {
                log.debug("Searching for the resource within repo:{}, path:{}", repo.getKey(),
                        context.getPath());
                BaseResource res = getResourceFromRepo.apply(context, repo);
                if (res != null && res.isFound() && !res.isBlocked()) {
                    log.debug("Resource was found in repo:{}, path:{}", repo.getKey(), context.getPath());
                    Map<String, LinkedHashSet<T>> extract =
                            controller.extract(res, context.getRepoKey(), context.getPath(), repo);
                    for (Map.Entry<String, LinkedHashSet<T>> entry : extract.entrySet()) {
                        String key = entry.getKey();
                        if (!rows.containsKey(key) || repo.isPriorityResolution() || !rows.get(key).isPriorities()) {
                            rows.computeIfAbsent(key,
                                    a -> new PackageMetadataVersionsSet<>(repo.isPriorityResolution()));
                            Set<T> metadata = rows.get(key).getVersions();
                            metadata.addAll(entry.getValue());
                        }
                    }
                } else {
                    forbidden = checkIfForbidden(res);
                    if (forbidden != null) {
                        log.debug("Request is forbidden by repo:{}, path:{}", repo.getKey(),
                                context.getPath());
                    }
                }
            }
            InputStream inputStream = controller.toInputStream(rows, context);
            return uploadService.uploadToVirtualOrRemoteCache(cachePath.getRepoKey(), cachePath.getPath(),
                    inputStream);
        } catch (PackageException e) {
            log.debug("Failed to persist artifact from virtual repo:{}, path:{}, reason:{}, error:{}",
                    cachePath.getRepoKey(),
                    cachePath.getPath(), e.getPublicReason(), e.getInternalReason());
            return new PackageUnfoundedResource(context.getRepoKey(), context.getPath(), e.getStatus(),
                    e.getPublicReason());
        }
    }

    /**
     * Iterate over a list of repos, and return a resource (if found) according to the VirtualResolverState
     * Important - if a resource will be found under a prioritised repository, we will not iterate over the regular
     * repositories.
     * Use this method in order to resolve files, or in order merge results from the sub repositories using
     * PackageVirtualResolver#overrideResource
     *
     * @param context             - original request context
     * @param resolver            - resolver that handling that resource if found
     * @param repositories        - sub repositories of a virtual repository - ordered by prioritised locals ->
     *                            prioritised remotes, locals -> remotes.
     * @param getResourceFromRepo BIFunction that get requestContext and real repo as parameters, and returning a
     *                            resource.
     */
    private BaseResource searchInRepositories(PackageRequestContext context,
            PackageVirtualResolver resolver, List<? extends PackageRepo> repositories,
            BiFunction<PackageRequestContext, PackageRepo, BaseResource> getResourceFromRepo) {
        // Save forbidden unfound response
        PackageUnfoundedResource forbidden = null;
        boolean foundByPriority = false;
        for (PackageRepo repo : repositories) {
            if (foundByPriority && !repo.isPriorityResolution()) {
                break;
            }
            log.debug("Searching for the resource within repo:{} '{}'", repo.getKey(), context.getRequestContextInfo());
            BaseResource res = getResourceFromRepo.apply(context, repo);
            if (res != null && res.isFound() && !res.isBlocked()) {
                String repoKey = repo.getKey();
                log.debug("Resource was found in repo:{} '{}'", repoKey, context.getRequestContextInfo());
                switch (resolver.getResolverState()) {
                    case returnCurrentValue: {
                        log.debug("Selecting: repo item from repo:{} '{}'", repoKey, context.getRequestContextInfo());
                        return res;
                    }
                    case skipCurrentValue: {
                        log.debug("Skipping: item from repo:{} '{}'", repoKey, context.getRequestContextInfo());
                        break;
                    }
                    case collected: {
                        foundByPriority = repo.isPriorityResolution();
                        log.debug("Collected: item from repo:{} '{}'", repoKey, context.getRequestContextInfo());
                        break;
                    }
                }
            } else {
                forbidden = checkIfForbidden(res);
                if (forbidden != null) {
                    log.debug("Request is forbidden by repo:{}", repo.getKey());
                }
            }
        }
        BaseResource resource = resolver.overrideResource(context.getRepoKey(), context);
        if (resource != null && resource.isFound()) {
            log.debug("Selecting: overridden by adapter repo:{}, '{}'} ", resource.getRepoKey(),
                    resource.getPath());
            return resource;
        }
        // Not found in any repo
        return newUnfoundResource(context.getRepoKey(), context.getPath(), forbidden);
    }

    private PackageUnfoundedResource checkIfForbidden(BaseResource resource) {
        if (resource instanceof PackageUnfoundedResource &&
                ((PackageUnfoundedResource) resource).getStatusCode() == SC_FORBIDDEN) {
            return (PackageUnfoundedResource) resource;
        }
        return null;
    }

    private BaseResource newUnfoundResource(String repoKey, String path, PackageUnfoundedResource forbidden) {
        if (forbidden != null) {
            log.debug("Returning a forbidden-unfound resource");
            return new PackageUnfoundedResource(repoKey, path, forbidden.getStatusCode(), forbidden.getReason());
        } else {
            log.debug("Returning an unfound resource");
            return new PackageUnfoundedResource(repoKey, path, SC_NOT_FOUND,
                    "Could not find resource");
        }
    }

    @NotNull
    private PackageUnfoundedResource newUnfoundResource(PackageRequestContext ctx) {
        return new PackageUnfoundedResource(ctx.getRepoKey(), ctx.getPath(), SC_NOT_FOUND, "Not found");
    }
}
