/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.storage.db.version;

import com.google.common.annotations.VisibleForTesting;
import lombok.extern.slf4j.Slf4j;
import org.artifactory.storage.db.conversion.version.v227.CreateDbSpecificSequence;
import org.artifactory.storage.db.conversion.version.v228.AlterIndexAfterConcurrentCreation;
import org.artifactory.storage.db.properties.dao.DbPropertiesDao;
import org.artifactory.storage.db.version.converter.LockingDBSqlConverter;
import org.artifactory.storage.db.version.converter.ScriptFileDbConverter;
import org.artifactory.storage.db.version.converter.SkippableScriptFileDbConverter;
import org.artifactory.version.ArtifactoryServiceVersion;
import org.artifactory.version.ArtifactoryServiceVersionProvider;
import org.artifactory.version.CompoundVersionDetails;
import org.jfrog.converter.DbConverter;
import org.jfrog.converter.DbVersion;
import org.jfrog.converter.model.ConverterExecutionType;
import org.jfrog.converter.predefined.StatementDbConverter;
import org.jfrog.storage.DbType;

import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static org.artifactory.storage.db.build.dao.BuildsDao.BUILDS_TABLE;
import static org.artifactory.storage.db.bundle.dao.ArtifactBundlesDao.ARTIFACT_BUNDLES_TABLE;
import static org.artifactory.storage.db.bundle.dao.ArtifactBundlesDao.BUNDLE_FILES_TABLE;
import static org.artifactory.storage.db.dao.locks.DbDistributeLocksDao.TABLE_DISTRIBUTED_LOCKS;
import static org.artifactory.storage.db.event.dao.EventsDao.EVENTS_TABLE;
import static org.artifactory.storage.db.event.dao.EventsDao.EVENTS_TEMP_TABLE;
import static org.artifactory.storage.db.event.dao.NodeEventPrioritiesDao.EVENT_PRIORITIES_TABLE;
import static org.artifactory.storage.db.event.dao.NodeEventsCursorDao.EVENTS_CURSOR_TABLE;
import static org.artifactory.storage.db.event.error.dao.NodeEventsErrorsDAO.TABLE_NODE_EVENTS_ERRORS;
import static org.artifactory.storage.db.fs.dao.JobsDao.JOBS_TABLE;
import static org.artifactory.storage.db.fs.dao.NodesDao.NODES_TABLE;
import static org.artifactory.storage.db.fs.dao.PropertiesDao.PROPS_TABLE;
import static org.artifactory.storage.db.fs.dao.TasksDao.TASKS_TABLE;
import static org.artifactory.storage.db.keys.dao.TrustedKeysDao.TRUSTED_KEYS_TABLE;
import static org.artifactory.storage.db.servers.dao.ArtifactoryServersDao.SERVERS_TABLE;
import static org.artifactory.version.ArtifactoryServiceVersionProvider.v600m023;
import static org.artifactory.version.ArtifactoryServiceVersionProvider.v610m002;
import static org.artifactory.version.ArtifactoryServiceVersionProvider.v6110m001;
import static org.artifactory.version.ArtifactoryServiceVersionProvider.v6120m079;
import static org.artifactory.version.ArtifactoryServiceVersionProvider.v6170m001;
import static org.artifactory.version.ArtifactoryServiceVersionProvider.v620;
import static org.artifactory.version.ArtifactoryServiceVersionProvider.v640m007;
import static org.artifactory.version.ArtifactoryServiceVersionProvider.v650m001;
import static org.artifactory.version.ArtifactoryServiceVersionProvider.v660m001;
import static org.artifactory.version.ArtifactoryServiceVersionProvider.v682m001;
import static org.artifactory.version.ArtifactoryServiceVersionProvider.v690m001;
import static org.artifactory.version.ArtifactoryServiceVersionProvider.v7110m033;
import static org.artifactory.version.ArtifactoryServiceVersionProvider.v7130m001;
import static org.artifactory.version.ArtifactoryServiceVersionProvider.v7170m040;
import static org.artifactory.version.ArtifactoryServiceVersionProvider.v7201m001;
import static org.artifactory.version.ArtifactoryServiceVersionProvider.v7230m012;
import static org.artifactory.version.ArtifactoryServiceVersionProvider.v731m007;
import static org.artifactory.version.ArtifactoryServiceVersionProvider.v732m017;
import static org.artifactory.version.ArtifactoryServiceVersionProvider.v734m001;
import static org.artifactory.version.ArtifactoryServiceVersionProvider.v7380;
import static org.artifactory.version.ArtifactoryServiceVersionProvider.v7400;
import static org.artifactory.version.ArtifactoryServiceVersionProvider.v740m081;
import static org.artifactory.version.ArtifactoryServiceVersionProvider.v750m145;
import static org.artifactory.version.ArtifactoryServiceVersionProvider.v760m001;
import static org.artifactory.version.ArtifactoryServiceVersionProvider.v770;
import static org.jfrog.converter.model.ConverterExecutionType.SEQ_ASYNC;
import static org.jfrog.converter.model.ConverterExecutionType.SEQ_SYNC;
import static org.jfrog.converter.util.ColumnDefinition.builder;
import static org.jfrog.converter.util.DbColumnType.BIGINT;
import static org.jfrog.converter.util.DbColumnType.CHAR;
import static org.jfrog.converter.util.DbColumnType.VARCHAR;
import static org.jfrog.converter.util.PreDefinedConditions.columnExists;
import static org.jfrog.converter.util.PreDefinedConditions.columnSizeNotMatch;
import static org.jfrog.converter.util.PreDefinedConditions.dbSpecific;
import static org.jfrog.converter.util.PreDefinedConditions.executeUnconditional;
import static org.jfrog.converter.util.PreDefinedConditions.indexNotExists;
import static org.jfrog.converter.util.PreDefinedConditions.tableExists;
import static org.jfrog.converter.util.PreDefinedConditions.tableNotExists;
import static org.jfrog.converter.util.PreDefinedConverters.addColumn;
import static org.jfrog.converter.util.PreDefinedConverters.addColumnDefaultValue;
import static org.jfrog.converter.util.PreDefinedConverters.addConstraintForeignKey;
import static org.jfrog.converter.util.PreDefinedConverters.addConstraintPrimaryKey;
import static org.jfrog.converter.util.PreDefinedConverters.alterColumnNullability;
import static org.jfrog.converter.util.PreDefinedConverters.alterColumnSize;
import static org.jfrog.converter.util.PreDefinedConverters.createIndex;
import static org.jfrog.converter.util.PreDefinedConverters.createUniqueIndex;
import static org.jfrog.converter.util.PreDefinedConverters.dropConstraint;
import static org.jfrog.converter.util.PreDefinedConverters.dropDefault;
import static org.jfrog.converter.util.PreDefinedConverters.dropIndex;
import static org.jfrog.converter.util.PreDefinedConverters.dropTable;
import static org.jfrog.converter.util.PreDefinedConverters.renameColumn;
import static org.jfrog.converter.util.PreDefinedConverters.renameTable;

/**
 * Artifactory specific implementation for DbVersion. Each entry holds an info for one db converter.
 *
 * @author Michael Akushsky
 * @author Volodymyr Shulga
 * @author shaharz
 * @see DbVersion
 */
@Slf4j
@SuppressWarnings("squid:S1192")
public enum ArtifactoryDBVersion implements DbVersion {
    v10(SEQ_SYNC, createIndex(BUNDLE_FILES_TABLE, "repo_path_bundle_files_idx", List.of("repo_path"),
            Map.of(DbType.MYSQL, List.of("repo_path(255)"), DbType.MARIADB, List.of("repo_path(255)"))), v600m023),
    v11(SEQ_SYNC, new StatementDbConverter(
            "update trusted_keys set alias = kid where alias is null;",
            "Set all alias to kid in trusted_keys if trusted_keys equals NULL",
            executeUnconditional()), v610m002),
    v12(SEQ_SYNC, alterColumnNullability(TRUSTED_KEYS_TABLE, "alias", "VARCHAR(255)", false), v610m002),
    v13(SEQ_SYNC, createUniqueIndex(TRUSTED_KEYS_TABLE, "trusted_keys_alias", List.of("alias")), v610m002),
    v14(SEQ_SYNC, dropConstraint(SERVERS_TABLE, "artifactory_servers_pk", dbSpecific(DbType.MSSQL)), v620),
    v15(SEQ_SYNC, alterColumnSize(SERVERS_TABLE, "server_id", 128), v620),
    v16(SEQ_SYNC, alterColumnNullability(SERVERS_TABLE, "server_id", "VARCHAR(128)", false), v620),
    v17(SEQ_SYNC, addConstraintPrimaryKey(SERVERS_TABLE, "artifactory_servers_pk", "server_id",
            dbSpecific(DbType.MSSQL)), v620),
    v18(SEQ_SYNC, new ScriptFileDbConverter("v18_ui_session", tableNotExists("UI_SESSION")), v620),
    v19(SEQ_SYNC, createIndex("UI_SESSION", "UI_SESSION_IX1", List.of("LAST_ACCESS_TIME")), v620),
    v20(SEQ_SYNC, new ScriptFileDbConverter("v20_ui_session_attributes",
            tableNotExists("UI_SESSION_ATTRIBUTES")), v620),
    v21(SEQ_SYNC, createIndex("UI_SESSION_ATTRIBUTES", "UI_SESSION_ATTRIBUTES_IX1", List.of("SESSION_ID")), v620),
    v22(SEQ_SYNC, addColumn(builder().tableName(ARTIFACT_BUNDLES_TABLE).columnName("type").type(VARCHAR).size(32)
            .notNull(true).defaultValue("TARGET").build()), v640m007),
    v23(SEQ_SYNC, addColumn(builder().tableName(ARTIFACT_BUNDLES_TABLE).columnName("storing_repo")
            .type(VARCHAR).size(64).build()), v640m007),
    v24(SEQ_SYNC, dropIndex(ARTIFACT_BUNDLES_TABLE, "name_version_idx", List.of("name")), v640m007),
    v25(SEQ_SYNC, createUniqueIndex(ARTIFACT_BUNDLES_TABLE, "name_version_idx",
            List.of("name", "version", "type")), v640m007),
    v26(SEQ_SYNC, new ScriptFileDbConverter("v26_node_props_index", dbSpecific(DbType.POSTGRESQL)), v650m001),
    v27(SEQ_SYNC, new ScriptFileDbConverter("v27_replication_errors", tableNotExists("replication_errors")), v660m001),
    v28(SEQ_SYNC, createIndex("replication_errors", "replication_errors_rep_key_idx",
            List.of("replication_key")), v660m001),
    v29(SEQ_SYNC, dropIndex(NODES_TABLE, "nodes_repo_path_checksum", List.of("repo_path_checksum"),
            dbSpecific(DbType.MYSQL, DbType.MARIADB)), v660m001),
    v30(SEQ_SYNC, createIndex(NODES_TABLE, "nodes_repo_path_checksum", List.of("repo_path_checksum"),
            dbSpecific(DbType.MYSQL, DbType.MARIADB)), v660m001),
    v31(SEQ_SYNC, addColumn(builder().tableName(BUNDLE_FILES_TABLE)
            .columnName("original_component_details").type(VARCHAR).size(1000).defaultValue(null).build()), v682m001),
    v32(SEQ_SYNC, new ScriptFileDbConverter("v32_node_event_cursor_table",
            tableNotExists("node_event_cursor")), v690m001),
    v33(SEQ_SYNC, new ScriptFileDbConverter("v33_node_event_priorities_table",
            tableNotExists(EVENT_PRIORITIES_TABLE)), v690m001),
    v34(SEQ_SYNC, new ScriptFileDbConverter("v34_migration_status_table",
            tableNotExists("migration_status")), v6110m001),
    v35(SEQ_SYNC, new ScriptFileDbConverter("v35_create_jobs_table", tableNotExists(JOBS_TABLE)), v6110m001),
    v36(SEQ_SYNC, createIndex(JOBS_TABLE, "jobs_index",
            List.of("started", "finished", "job_status", "job_type")), v6110m001),
    v37(SEQ_SYNC, new StatementDbConverter(
            "delete from migration_status where identifier = 'metadata-service-migration';",
            "Delete all from 'migration_status' where identifier equals 'metadata-service-migration'",
            executeUnconditional()), v6110m001),
    v38(SEQ_SYNC, new StatementDbConverter(
            "update artifact_bundles set storing_repo = 'release-bundles' where type = 'TARGET';",
            "Set all storing_repo to 'release-bundles' in artifact_bundles if type equals 'TARGET'",
            executeUnconditional()), v6120m079),
    v39(SEQ_SYNC, dropIndex(ARTIFACT_BUNDLES_TABLE, "name_version_status_idx",
            List.of("name", "version", "status")), v6120m079),
    v40(SEQ_SYNC, createIndex(ARTIFACT_BUNDLES_TABLE, "name_ver_sta_date_typ_repo_idx",
            List.of("name", "version", "status", "date_created", "type", "storing_repo")), v6120m079),
    v41(SEQ_SYNC, createIndex(BUNDLE_FILES_TABLE, "bundle_id_bundle_files_idx", List.of("bundle_id")), v6120m079),
    v42(SEQ_SYNC, addColumn(builder().tableName(EVENTS_CURSOR_TABLE).columnName("type").type(VARCHAR)
            .size(255).notNull(true).defaultValue("UNDEFINED").build()), v6170m001),
    v43(SEQ_SYNC, new StatementDbConverter(
            "UPDATE node_event_cursor SET operator_id = 'metadata-operator-events', type = 'METADATA_EVENTS' " +
                    "WHERE operator_id = 'metadata-operator';",
            "Set `METADATA_EVENT` type for node_event_cursor records with `metadata-operator` id.",
            executeUnconditional()), v6170m001),
    v44(SEQ_SYNC, alterColumnSize(TASKS_TABLE, "task_context", 2048), v732m017),
    v45(SEQ_SYNC, new StatementDbConverter(
            "CREATE INDEX node_props_prop_key_value_idx ON node_props (prop_key) INCLUDE(prop_value);",
            "Create specific MSSQL index 'node_props_prop_key_value_idx' for 'prop_key' include 'prop_value' on " +
                    "'node_props'",
            dbSpecific(DbType.MSSQL),
            indexNotExists(PROPS_TABLE, "node_props_prop_key_value_idx", List.of("prop_key"))), v740m081),
    v46(SEQ_SYNC, new StatementDbConverter(
            "CREATE INDEX node_props_prop_value_key_idx ON node_props (prop_value) INCLUDE(prop_key);",
            "Create specific MSSQL index 'node_props_prop_value_key_idx' for 'prop_value' include 'prop_key' on " +
                    "'node_props'",
            dbSpecific(DbType.MSSQL),
            indexNotExists(PROPS_TABLE, "node_props_prop_value_key_idx", List.of("prop_value"))), v740m081),
    v47(SEQ_SYNC, new StatementDbConverter(
            "CREATE INDEX node_props_node_prop_value_idx ON node_props(node_id) INCLUDE(prop_key, prop_value);",
            "Create specific MSSQL index 'node_props_node_prop_value_idx' for 'node_id' include 'prop_key' and " +
                    "'prop_value' on 'node_props'",
            dbSpecific(DbType.MSSQL),
            indexNotExists(PROPS_TABLE, "node_props_node_prop_value_idx", List.of("node_id"))), v740m081),
    v48(SEQ_SYNC, dropIndex(PROPS_TABLE, "node_props_node_id_idx", List.of("node_id"),
            dbSpecific(DbType.MSSQL)), v740m081),
    v49(SEQ_SYNC, dropIndex(PROPS_TABLE, "node_props_prop_key_idx", List.of("prop_key"),
            dbSpecific(DbType.MSSQL)), v740m081),
    v50(SEQ_SYNC,
            dropIndex(PROPS_TABLE, "node_props_prop_value_idx", List.of("prop_value"),
                    dbSpecific(DbType.MSSQL)), v740m081),
    v51(SEQ_SYNC, alterColumnSize(ARTIFACT_BUNDLES_TABLE, "version", 32), v750m145),
    v52(SEQ_SYNC, alterColumnSize(SERVERS_TABLE, "artifactory_version", 128), v760m001),
    v53(SEQ_SYNC, alterColumnNullability(SERVERS_TABLE, "artifactory_version", "VARCHAR(128)", false,
            dbSpecific(DbType.MYSQL, DbType.MARIADB, DbType.MSSQL)), v760m001),
    v54(SEQ_SYNC, alterColumnSize(DbPropertiesDao.TABLE_NAME, "artifactory_version", 128), v760m001),
    v55(SEQ_SYNC, alterColumnNullability(DbPropertiesDao.TABLE_NAME, "artifactory_version", "VARCHAR(128)", false,
            dbSpecific(DbType.MYSQL, DbType.MARIADB, DbType.MSSQL)), v760m001),
    v56(SEQ_SYNC, renameColumn("replication_errors", "replication_key", "operation_key", "VARCHAR(255)"), v770),
    v57(SEQ_SYNC, renameColumn("replication_errors", "task_time", "event_time", "BIGINT NOT NULL"), v770),
    v58(SEQ_SYNC, renameColumn("replication_errors", "task_type", "event_type", "SMALLINT NOT NULL"), v770),
    v59(SEQ_SYNC, renameColumn("replication_errors", "task_path", "path", "VARCHAR(1344) NOT NULL"), v770),
    v60(SEQ_SYNC, addColumn(builder().tableName("replication_errors").columnName("event_category")
            .type(VARCHAR).size(225).build()), v770),
    v61(SEQ_SYNC, alterColumnNullability("replication_errors", "operation_key", "VARCHAR(255)", true), v770),
    v62(SEQ_SYNC, renameTable("replication_errors", "node_events_errors"), v770),
    v63(SEQ_SYNC, new StatementDbConverter(
            "UPDATE node_events_errors SET event_category = 'REPLICATION' WHERE event_category IS NULL;",
            "Set all event_category to 'REPLICATION' in node_events_errors if event_category equals NULL",
            tableExists(TABLE_NODE_EVENTS_ERRORS)), v770),
    v64(SEQ_SYNC, alterColumnNullability("node_events_errors", "event_category", "VARCHAR(255)", false,
            indexNotExists(TABLE_NODE_EVENTS_ERRORS, "node_events_errors_idx",
                    List.of("operation_key", "event_category", "event_id", "event_time"))), v770),
    // failed skipped not changed to skipped??
    v65(SEQ_SYNC, new StatementDbConverter(
            "UPDATE node_events_errors SET event_type = 1 WHERE event_type = 2;",
            "Set all event_type to '1' in node_events_errors if event_type equals '2'",
            columnExists(TABLE_NODE_EVENTS_ERRORS, "event_type")), v770),
    v66(SEQ_SYNC, dropIndex(TABLE_NODE_EVENTS_ERRORS, "replication_errors_rep_key_idx", List.of("operation_key")),
            v770),
    v67(SEQ_SYNC, createIndex(TABLE_NODE_EVENTS_ERRORS, "node_events_errors_op_key_idx", List.of("operation_key")),
            v770),
    v68(SEQ_SYNC, alterColumnSize(SERVERS_TABLE, "server_role", 255), v770),
    v69(SEQ_SYNC, alterColumnNullability(SERVERS_TABLE, "server_role", "VARCHAR(255)", false,
            dbSpecific(DbType.MYSQL, DbType.MARIADB, DbType.MSSQL)), v770),
    v70(SEQ_SYNC, dropTable("aces"), v770),
    v71(SEQ_SYNC, dropTable("acls"), v770),
    v72(SEQ_SYNC, dropTable("permission_target_repos"), v770),
    v73(SEQ_SYNC, dropTable("permission_targets"), v770),
    v74(SEQ_SYNC, dropTable("users_groups"), v770),
    v75(SEQ_SYNC, dropTable("groups"), v770),
    v76(SEQ_SYNC, dropTable("user_props"), v770),
    v77(SEQ_SYNC, dropTable("users"), v770),
    v78(SEQ_ASYNC, new ScriptFileDbConverter("v78_change_nodes_node_name_idx",
            dbSpecific(DbType.POSTGRESQL), indexNotExists(NODES_TABLE, "nodes_node_name_idx",
            List.of("node_name varchar_pattern_ops"))), v7110m033),
    v79(SEQ_ASYNC, new ScriptFileDbConverter("v79_change_nodes_node_path_idx", dbSpecific(DbType.POSTGRESQL),
            indexNotExists(NODES_TABLE, "nodes_node_path_idx", List.of("node_path varchar_pattern_ops"))), v7110m033),
    v80(SEQ_ASYNC, new ScriptFileDbConverter("v80_change_nodes_node_repo_path_idx", dbSpecific(DbType.POSTGRESQL),
            indexNotExists(NODES_TABLE, "nodes_repo_path_name_idx", List.of("repo varchar_pattern_ops",
                    "node_path varchar_pattern_ops", "node_name varchar_pattern_ops"))),
            v7110m033),
    v81(SEQ_ASYNC, createIndex("bundle_files", "bundle_files_node_id_idx", List.of("node_id")), v7110m033),
    v82(SEQ_SYNC, new LockingDBSqlConverter("v82_alter_distributed_locks_set_unlogged",
            dbSpecific(DbType.POSTGRESQL)), v7110m033),
    v83(SEQ_SYNC, addColumn(builder().tableName(BUILDS_TABLE).columnName("repo").type(VARCHAR)
            .size(64).notNull(true).defaultValue("artifactory-build-info").build()), v7130m001),
    v84(SEQ_SYNC, dropIndex(BUILDS_TABLE, "builds_name_number_date_idx", List.of("build_name")), v7130m001),
    v85(SEQ_SYNC, createUniqueIndex(BUILDS_TABLE, "builds_name_num_date_repo_idx",
            List.of("build_name", "build_number", "build_date", "repo")), v7130m001),
    v86(SEQ_SYNC, new CreateDbSpecificSequence(), v7170m040),
    v87(SEQ_SYNC, new ScriptFileDbConverter("v87_node_events_tmp_table", tableNotExists(EVENTS_TEMP_TABLE)), v7170m040),
    v88(SEQ_SYNC, createIndex(EVENTS_TEMP_TABLE, "node_events_tmp_idx",
            List.of("timestamp", "tmp_event_id")), v7170m040),
    v89(SEQ_SYNC, addColumn(builder().tableName(EVENTS_TABLE).columnName("hints").type(VARCHAR)
            .size(4000).build()), v7170m040),
    v90(SEQ_SYNC, addColumn(builder().tableName(TABLE_NODE_EVENTS_ERRORS).columnName("event_id")
            .type(BIGINT).build()), v7170m040),
    v91(SEQ_SYNC,
            addConstraintForeignKey(TABLE_NODE_EVENTS_ERRORS, "node_events_errors_event_id_fk", "event_id",
                    EVENTS_TABLE, "event_id"), v7170m040),
    v92(SEQ_SYNC, dropIndex(TABLE_NODE_EVENTS_ERRORS, "node_events_errors_op_key_idx", List.of("operation_key")),
            v7170m040),
    v93(SEQ_SYNC, createIndex(TABLE_NODE_EVENTS_ERRORS, "node_events_errors_idx",
            List.of("operation_key", "event_category", "event_id", "event_time")), v7170m040),
    v94(SEQ_SYNC, addColumn(builder().tableName("node_meta_infos").columnName("props_md5").type(CHAR).size(32)
            .build()), v7170m040),
    v95(SEQ_ASYNC, createIndex(EVENTS_TEMP_TABLE, "node_events_tmp_event_id_idx", List.of("tmp_event_id")), v7201m001),
    v96(SEQ_SYNC, dropIndex(TABLE_NODE_EVENTS_ERRORS, "node_events_errors_idx",
            List.of("operation_key", "event_category", "event_id", "event_time"), dbSpecific(DbType.MSSQL)), v7201m001),
    v97(SEQ_SYNC, new StatementDbConverter(
            "UPDATE node_events_errors SET event_category = 'REPLICATION' WHERE event_category IS NULL;",
            "Set all events in `node_events_errors` missing category to `REPLICATION`",
            dbSpecific(DbType.MSSQL)), v7201m001),
    v98(SEQ_SYNC, alterColumnNullability(TABLE_NODE_EVENTS_ERRORS, "event_category", "VARCHAR(255)", false,
            columnSizeNotMatch(TABLE_NODE_EVENTS_ERRORS, "event_category", 255),
            dbSpecific(DbType.MSSQL)), v7201m001),
    v99(SEQ_SYNC, createIndex(TABLE_NODE_EVENTS_ERRORS, "node_events_errors_idx",
            List.of("operation_key", "event_category", "event_id", "event_time"), dbSpecific(DbType.MSSQL)), v7201m001),
    v100(SEQ_SYNC, addColumn(builder().tableName("build_modules").columnName("type").type(VARCHAR)
            .size(64).defaultValue(null).build()), v7230m012),
    v101(SEQ_SYNC, alterColumnSize("watches", "username", 255)),
    v102(SEQ_SYNC, addColumn(builder().tableName("build_modules").columnName("type").type(VARCHAR).size(64)
            .defaultValue(null).build())),
    v103(SEQ_SYNC, new ScriptFileDbConverter("v103_binaries_tasks", tableNotExists("binaries_tasks"))),
    v104(SEQ_SYNC, createIndex("binaries_tasks", "binaries_tasks_time_idx", List.of("time"))),
    v105(SEQ_SYNC, createIndex("binaries_tasks", "binaries_tasks_start_time_idx", List.of("start_time"))),
    v106(SEQ_SYNC, createIndex("binaries_tasks", "binaries_tasks_server_id_idx", List.of("server_id"))),
    v107(SEQ_SYNC, createIndex("binaries_tasks", "binaries_tasks_count_idx", List.of("count"))),
    v108(SEQ_SYNC, alterColumnSize("node_props", "prop_value", 4000)),
    v109(SEQ_SYNC,
            new ScriptFileDbConverter("v109_add_distributed_locks_lock_key_idx", tableExists(TABLE_DISTRIBUTED_LOCKS),
                    indexNotExists(TABLE_DISTRIBUTED_LOCKS, "lock_key_idx", List.of("lock_key")))),
    v110(SEQ_SYNC,
            addColumn(builder().tableName(ARTIFACT_BUNDLES_TABLE).columnName("date_distributed").type(BIGINT).build())),
    v111(SEQ_ASYNC, new ScriptFileDbConverter("v111_prepare_change_node_props_node_prop_value_idx",
            dbSpecific(DbType.POSTGRESQL), indexNotExists("node_props", "pattern_node_props_node_prop_value_idx",
            List.of("node_id", "prop_key varchar_pattern_ops", "prop_value varchar_pattern_ops"))), v731m007),
    v112(SEQ_ASYNC, new AlterIndexAfterConcurrentCreation("node_props_node_prop_value_idx",
            "pattern_node_props_node_prop_value_idx", "v112_alter_node_props_node_prop_value_idx",
            dbSpecific(DbType.POSTGRESQL), indexNotExists("node_props", "node_props_node_prop_value_idx",
            List.of("node_id", "prop_key varchar_pattern_ops", "prop_value varchar_pattern_ops"))), v731m007),
    v113(SEQ_ASYNC, new ScriptFileDbConverter("v113_prepare_change_node_props_prop_key_value_idx",
            dbSpecific(DbType.POSTGRESQL), indexNotExists("node_props", "pattern_node_props_prop_key_value_idx",
            List.of("prop_key varchar_pattern_ops", "prop_value varchar_pattern_ops"))), v731m007),
    v114(SEQ_ASYNC, new AlterIndexAfterConcurrentCreation("node_props_prop_key_value_idx",
            "pattern_node_props_prop_key_value_idx", "v114_alter_node_props_prop_key_value_idx",
            dbSpecific(DbType.POSTGRESQL), indexNotExists("node_props", "node_props_prop_key_value_idx",
            List.of("prop_key varchar_pattern_ops", "prop_value varchar_pattern_ops"))), v731m007),
    v115(SEQ_ASYNC, new ScriptFileDbConverter("v115_prepare_change_node_props_prop_value_key_idx",
            dbSpecific(DbType.POSTGRESQL), indexNotExists("node_props", "pattern_node_props_prop_value_key_idx",
            List.of("prop_value varchar_pattern_ops", "prop_key varchar_pattern_ops"))), v731m007),
    v116(SEQ_ASYNC, new AlterIndexAfterConcurrentCreation("node_props_prop_value_key_idx",
            "pattern_node_props_prop_value_key_idx", "v116_alter_node_props_prop_value_key_idx",
            dbSpecific(DbType.POSTGRESQL), indexNotExists("node_props", "node_props_prop_value_key_idx",
            List.of("prop_value varchar_pattern_ops", "prop_key varchar_pattern_ops"))), v731m007),
    v117(SEQ_SYNC, addColumnDefaultValue(ARTIFACT_BUNDLES_TABLE, "type", "'TARGET'")),
    v118(SEQ_SYNC,
            addColumn(builder().tableName(EVENTS_CURSOR_TABLE).columnName("status")
                    .type(VARCHAR).size(255).defaultValue("SYNC").build())),
    v119(SEQ_SYNC,
            addColumn(builder().tableName(EVENTS_CURSOR_TABLE).columnName("last_updated")
                    .type(BIGINT).defaultValue(System.currentTimeMillis()).build())),
    v120(SEQ_ASYNC, new SkippableScriptFileDbConverter(
            new ScriptFileDbConverter("v120_prepare_change_node_props_node_prop_value_idx",
                    dbSpecific(DbType.POSTGRESQL),
                    indexNotExists("node_props", "pattern_node_props_node_prop_value_idx",
                            List.of("node_id", "prop_key varchar_pattern_ops", "prop_value varchar_pattern_ops")))),
            v734m001),
    v121(SEQ_ASYNC,
            new SkippableScriptFileDbConverter(new AlterIndexAfterConcurrentCreation("node_props_node_prop_value_idx",
                    "pattern_node_props_node_prop_value_idx", "v121_alter_node_props_node_prop_value_idx",
                    dbSpecific(DbType.POSTGRESQL), indexNotExists("node_props", "node_props_node_prop_value_idx",
                    List.of("node_id", "prop_key varchar_pattern_ops", "prop_value varchar_pattern_ops")))), v734m001),
    v122(SEQ_ASYNC, new SkippableScriptFileDbConverter(
            new ScriptFileDbConverter("v122_prepare_change_node_props_prop_key_value_idx",
                    dbSpecific(DbType.POSTGRESQL), indexNotExists("node_props", "pattern_node_props_prop_key_value_idx",
                    List.of("prop_key varchar_pattern_ops", "prop_value varchar_pattern_ops")))), v734m001),
    v123(SEQ_ASYNC,
            new SkippableScriptFileDbConverter(new AlterIndexAfterConcurrentCreation("node_props_prop_key_value_idx",
                    "pattern_node_props_prop_key_value_idx", "v123_alter_node_props_prop_key_value_idx",
                    dbSpecific(DbType.POSTGRESQL), indexNotExists("node_props", "node_props_prop_key_value_idx",
                    List.of("prop_key varchar_pattern_ops", "prop_value varchar_pattern_ops")))), v734m001),
    v124(SEQ_ASYNC, new SkippableScriptFileDbConverter(
            new ScriptFileDbConverter("v124_prepare_change_node_props_prop_value_key_idx",
                    dbSpecific(DbType.POSTGRESQL), indexNotExists("node_props", "pattern_node_props_prop_value_key_idx",
                    List.of("prop_value varchar_pattern_ops", "prop_key varchar_pattern_ops")))), v734m001),
    v125(SEQ_ASYNC,
            new SkippableScriptFileDbConverter(new AlterIndexAfterConcurrentCreation("node_props_prop_value_key_idx",
                    "pattern_node_props_prop_value_key_idx", "v125_alter_node_props_prop_value_key_idx",
                    dbSpecific(DbType.POSTGRESQL), indexNotExists("node_props", "node_props_prop_value_key_idx",
                    List.of("prop_value varchar_pattern_ops", "prop_key varchar_pattern_ops")))), v734m001),
    v126(SEQ_SYNC, dropDefault(EVENTS_CURSOR_TABLE, "status"), v7380),
    v127(SEQ_SYNC, dropDefault(EVENTS_CURSOR_TABLE, "last_updated"), v7380),
    v128(SEQ_SYNC, alterColumnSize("distributed_locks", "category", 128,
            dbSpecific(DbType.DERBY, DbType.MARIADB, DbType.MYSQL, DbType.POSTGRESQL,
                    DbType.ORACLE)), v7400),
    v129(SEQ_SYNC, new ScriptFileDbConverter("v129_alter_category_size_column",
            dbSpecific(DbType.MSSQL)), v7400);

    private final ConverterExecutionType type;
    private final DbConverter converter;
    @VisibleForTesting
    final ArtifactoryServiceVersionProvider applyFromVersion;

    /**
     * Constructor that should be used for all newly introduced DB converters (starting from v101)
     *
     * @param type      - converter type. Used to filter out and further process converters of a same type.
     * @param converter - conditional converter containing actual work.
     */
    ArtifactoryDBVersion(ConverterExecutionType type, DbConverter converter) {
        this(type, converter, null);
    }

    /**
     * Constructor used for converters that were executed by the moment of new DB converters execution model
     * introducing (prior v101).
     *
     * @param type               - converter type. Used to filter out and further process converters of a same type.
     * @param converter          - conditional converter containing actual work.
     * @param artifactoryVersion - version of artifactory for which this converter is actual. If this prop is less than
     *                           current artifactory version, this converter will be skipped.
     */
    ArtifactoryDBVersion(ConverterExecutionType type, DbConverter converter,
            ArtifactoryServiceVersionProvider artifactoryVersion) {
        this.type = type;
        this.converter = converter;
        this.applyFromVersion = artifactoryVersion;
    }

    public String version() {
        return this.name();
    }

    public String description() {
        return converter.description();
    }

    public ConverterExecutionType type() {
        return this.type;
    }

    public DbConverter converter() {
        return this.converter;
    }

    public boolean isAlreadyApplied(CompoundVersionDetails sourceArtifactory) {
        return applyFromVersion != null && sourceArtifactory.getVersion().after(applyFromVersion.get());
    }

    public String systemVersion() {
        return Optional.ofNullable(applyFromVersion)
                .map(ArtifactoryServiceVersionProvider::get)
                .map(ArtifactoryServiceVersion::getVersion)
                .orElse("");
    }

    public static DbVersion[] get(ConverterExecutionType... execTypes) {
        Set<ConverterExecutionType> filter = execTypes.length > 0 ?
                Stream.of(execTypes).collect(Collectors.toSet()) :
                Stream.of(ConverterExecutionType.values()).collect(Collectors.toSet());

        return Stream.of(ArtifactoryDBVersion.values())
                .filter(version -> filter.contains(version.type()))
                .sorted(Comparator.comparing(ArtifactoryDBVersion::ordinal))
                .toArray(DbVersion[]::new);
    }

    public static DbVersion[] getAlreadyApplied(CompoundVersionDetails sourceArtifactory,
            ConverterExecutionType... execTypes) {
        Set<ConverterExecutionType> filter = execTypes.length > 0 ?
                Stream.of(execTypes).collect(Collectors.toSet()) :
                Stream.of(ConverterExecutionType.values()).collect(Collectors.toSet());

        return Stream.of(ArtifactoryDBVersion.values())
                .filter(version -> filter.contains(version.type()))
                .filter(version -> version.isAlreadyApplied(sourceArtifactory))
                .sorted(Comparator.comparing(ArtifactoryDBVersion::ordinal))
                .toArray(DbVersion[]::new);
    }
}

