package org.artifactory.rest.filter;

import com.google.common.annotations.VisibleForTesting;
import lombok.extern.slf4j.Slf4j;
import org.artifactory.aol.AolService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.security.AuthenticationHelper;
import org.artifactory.security.access.AccessTokenAuthentication;
import org.glassfish.jersey.server.model.AnnotatedMethod;
import org.jfrog.access.common.PermissionScope;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import javax.annotation.Priority;
import javax.annotation.security.RolesAllowed;
import javax.ws.rs.Priorities;
import javax.ws.rs.container.ContainerRequestContext;
import javax.ws.rs.container.ContainerRequestFilter;
import javax.ws.rs.container.DynamicFeature;
import javax.ws.rs.container.ResourceInfo;
import javax.ws.rs.core.FeatureContext;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.SecurityContext;
import javax.ws.rs.ext.Provider;
import java.io.IOException;
import java.security.Principal;

import static org.jfrog.common.ArgUtils.requireNonNull;

/**
 * @author Oleksandr Legenkiy
 */
@Component
@Provider
@Slf4j
public class AllowedPermissionScopeDynamicFeature implements DynamicFeature {

    private static final Class<AllowedPermissionScope> ANNOTATION_CLASS = AllowedPermissionScope.class;
    private final AuthorizationService authorizationService;
    private final AolService aolService;

    @Autowired
    public AllowedPermissionScopeDynamicFeature(AuthorizationService authorizationService, AolService aolService) {
        this.authorizationService = authorizationService;
        this.aolService = aolService;
    }

    @Override
    public void configure(ResourceInfo resourceInfo, FeatureContext context) {
        AnnotatedMethod am = new AnnotatedMethod(resourceInfo.getResourceMethod());
        AllowedPermissionScope allowedPermissionScope = null;
        if (am.isAnnotationPresent(ANNOTATION_CLASS)) {
            allowedPermissionScope = am.getAnnotation(ANNOTATION_CLASS);
        } else if (resourceInfo.getResourceClass().isAnnotationPresent(ANNOTATION_CLASS)) {
            allowedPermissionScope = resourceInfo.getResourceClass().getAnnotation(ANNOTATION_CLASS);
        }
        if (allowedPermissionScope != null) {
            PermissionScope permissionScope;
            try {
                permissionScope = PermissionScope.parse(allowedPermissionScope.value());
            } catch (Exception e) {
                log.error("Failed to parse permission scope: '{}'. {}", allowedPermissionScope.value(), e.getMessage());
                throw e;
            }
            context.register(new PermissionScopeFilter(authorizationService, permissionScope,
                    !am.isAnnotationPresent(RolesAllowed.class) &&
                            !resourceInfo.getResourceClass().isAnnotationPresent(RolesAllowed.class), aolService));
        }
    }

    // Should be called before the authorization filters because this filter can update the security context
    @Priority(Priorities.AUTHORIZATION - 100)
    @VisibleForTesting
    static class PermissionScopeFilter implements ContainerRequestFilter {

        private final AuthorizationService authorizationService;
        private final PermissionScope permissionScope;
        private final boolean shouldAbortNotAuthorized;
        private final AolService aolService;

        @VisibleForTesting
        PermissionScopeFilter(AuthorizationService authorizationService, PermissionScope permissionScope,
                boolean shouldAbortNotAuthorized, AolService aolService) {
            this.authorizationService =
                    requireNonNull(authorizationService, "authorizationService is required for the filter");
            this.permissionScope = requireNonNull(permissionScope, "permission scope is required for the filter");
            this.shouldAbortNotAuthorized = shouldAbortNotAuthorized;
            this.aolService = aolService;
        }

        @Override
        public void filter(ContainerRequestContext requestContext) throws IOException {
            if (isPermissionGranted() || isDashboardUser()) {
                // Make it ignore any @RolesAllowed value specified for the method
                requestContext.setSecurityContext(new AllRolesAuthenticator(authorizationService.currentUsername()));
            } else if (shouldAbortNotAuthorized) {
                requestContext.abortWith(Response.status(Response.Status.FORBIDDEN).build());
            }
        }

        private boolean isDashboardUser() {
            return aolService.isAol() && aolService.isDashboardUser();
        }

        private boolean isPermissionGranted() {
            return AuthenticationHelper.getAuthentication() instanceof AccessTokenAuthentication &&
                    authorizationService.isGrantedByAccessToken(permissionScope);
        }
    }


    @VisibleForTesting
    static class AllRolesAuthenticator implements SecurityContext {

        private final Principal principal;

        @VisibleForTesting
        AllRolesAuthenticator(String name) {
            this.principal = () -> name;
        }

        @Override
        public Principal getUserPrincipal() {
            return principal;
        }

        @Override
        public boolean isUserInRole(String role) {
            return true;
        }

        @Override
        public boolean isSecure() {
            return false;
        }

        @Override
        public String getAuthenticationScheme() {
            return SecurityContext.BASIC_AUTH;
        }
    }
}