package org.artifactory.rest.resource.system;

import org.artifactory.aol.AolService;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.mail.MailService;
import org.artifactory.api.rest.mail.MailMessageRequest;
import org.artifactory.rest.resource.email.EmailResource;
import org.artifactory.util.EmailException;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import javax.servlet.http.HttpServletResponse;
import static org.testng.Assert.assertEquals;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.doThrow;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;


@Test
public class MailResourceTest {

    @Mock
    private MailService mailService;

    @Mock
    private AolService aolService;

    private AutoCloseable autoCloseable;

    @BeforeMethod
    public void beforeMethod() throws Exception {
        autoCloseable = MockitoAnnotations.openMocks(this);
        ArtifactoryContext context = mock(ArtifactoryContext.class);
        when(context.beanForType(AolService.class)).thenReturn(aolService);
        ArtifactoryContextThreadBinder.bind(context);
    }

    @AfterMethod
    void tearDown() throws Exception {
        if (autoCloseable != null) {
            autoCloseable.close();
            autoCloseable = null;
        }
        ArtifactoryContextThreadBinder.unbind();
    }

    @DataProvider(name = "mailData")
    public Object[][] mailData() {
        return new Object[][] {
                {"to", "subject", "body", true, HttpServletResponse.SC_METHOD_NOT_ALLOWED},
                {"a@a.com", "1subject1", "body", false, HttpServletResponse.SC_OK},
        };
    }

    @Test
    void testSendMail() {
        EmailResource mailResource = new EmailResource(mailService);
        doNothing().when(mailService).sendMail(any(), eq("subject"), any());
        MailMessageRequest mailReq = new MailMessageRequest("a@a.com", "subject", "body");
        assertEquals(mailResource.sendMail(mailReq).getStatus(), HttpServletResponse.SC_OK);
        verify(mailService, times(1)).sendMail(new String[] {"a@a.com"}, "subject", "body");
    }

    @Test
    void testSendMailFail() {
        EmailResource mailResource = new EmailResource(mailService);
        doThrow(new EmailException("err")).when(mailService).sendMail(any(), eq("1subject1"), any());
        MailMessageRequest mailReq = new MailMessageRequest("a@a.com", "1subject1", "body");
        assertEquals(mailResource.sendMail(mailReq).getStatus(),HttpServletResponse.SC_BAD_REQUEST);
        verify(mailService, times(1)).sendMail(new String[] {"a@a.com"}, "1subject1", "body");
    }

    @Test(dataProvider = "mailData")
    void testSendMailNotAllowedOnSaaS(String recipient, String subject, String body, boolean onSaaS,
            int expectedResponseStatus) {
        when(aolService.isAol()).thenReturn(onSaaS);
        EmailResource mailResource = new EmailResource(mailService);
        MailMessageRequest mailReq = new MailMessageRequest(recipient, subject, body);
        assertEquals(expectedResponseStatus, mailResource.sendMail(mailReq).getStatus());
    }
}